

class TheanoToken():
    """
    A enum-like class representing mappings from string values to theano types 
    """
    int_vector = "VectorI"
    float_vector = "VectorF"
    int_scalar = "int"
    float_scalar = "float"
    int_matrix = "MatrixI"
    float_matrix = "MatrixD"
    single_element_float_vector = 'Single_element_VectorF'
    single_element_int_vector = 'Single_element_VectorI'


    
